// XPointer.java
// Justin Ludiwg 1999.03.01

package edu.wooster.lu99jp26.link;

import com.docuverse.dom.NodeListImpl;
import edu.wooster.lu99jp26.util.Function;
import edu.wooster.lu99jp26.util.VolatileInteger;
import java.lang.Integer;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * manipulates and resolves XPointers
 * 
 * id() only works with attribute names id or ID
 *
 * span() uses only the first node in the node list returned
 * from the XPointers of the first and second parameters
 * and the first node must come before second in the document
 *
 * implemented as a thread which blocks callers to getNewSelection()
 * until it resolves the XPointer, using wait() and notifyAll().
 * the XPointer thread protects the calling thread
 * if the virtual machine throws a StackOverflowException,
 * since many of these routines rely heavily on recursion
 */
public class XPointer
{
	protected Document doc;
	protected Node origin;
	protected String expr;
	protected NodeList selection = null;
	
	
	/**
	 * strips xpointer expression & connector from a URL
	 * @param url String the url
	 * @return String the url minus the xpointer expression & connector
	 */
	public static String stripXPointer (String url)
	{
		if (url.indexOf ('#') >= 0)
			return url.substring (0, url.indexOf ('#'));
		
		else if (url.indexOf ('|') >= 0)
			return url.substring (0, url.indexOf ('|'));
		
		return url;
	}
	
	/**
	 * gets the xpointer expression from a URL
	 * @param url String the url
	 * @return String the xpointer expression
	 */
	public static String getXPointer (String url)
	{
		if (url.indexOf ('#') >= 0)
			return url.substring (url.indexOf ('#') + 1);
		
		else if (url.indexOf ('|') >= 0)
			return url.substring (url.indexOf ('|') + 1);
		
		return "";
	}
	
	/**
	 * gets the next term of an XPointer expression
	 * @param expr String the XPointer expression
	 * @return String the next term
	 */
	public static String getNextTerm (String expr)
	{
		if (expr.indexOf ('.') > 0)
			return expr.substring (0, expr.indexOf ('.'));
		
		return expr;
	}
	
	/**
	 * strips the next term from an XPointer expression
	 * @param expr String the XPointer expression
	 * @return String the XPointer expression minus the next term
	 */
	public static String stripNextTerm (String expr)
	{
		if (expr.indexOf ('.') > 0)
			return expr.substring (expr.indexOf ('.') + 1);
		
		return "";
	}
	
	
	/**
	 * creates new XPointer
	 * @param doc Document the DOM Document into which this XPointer points
	 * @param origin Node the Node in which this XPointer resides
	 *    note this could be from a different Document than doc
	 * @param expr String the XPointer expression
	 */
	public XPointer (Document doc, Node origin, String expr)
	{
		super ();
		
		this.doc = doc;
		this.origin = origin;
		this.expr = expr;
	}
	
	/**
	 * gets the list of nodes this XPointer selects
	 * @return NodeList the list of nodes (may be empty)
	 */
	public NodeList getSelection ()
	{
		if (selection == null)
			selection = getNewSelection ();
		
		return selection;
	}
	
	/**
	 * gets the list of nodes this XPointer selects
	 * re-evaluates the XPointer expression if it's already been evaluated
	 * @return NodeList the list of nodes (may be empty)
	 */
	public NodeList getNewSelection ()
	{
		selection = resolve (doc.getDocumentElement (), expr);
		return selection;
	}
	
	/**
	 * resolves each term in expr, starting with Node start
	 * some relative terms may result in multiple nodes
	 * so all relative terms may apply to multiple source nodes
	 */
	protected NodeListImpl resolve (Node start, String expr)
	{
		// strip & parse each term
		expr = new String (expr);
		
		NodeListImpl result = new NodeListImpl (1);
		
		if (start != null)
			result.appendNode (start);
		
		// only continue if last term had result
		while (!expr.equals ("") && result.getLength () > 0)
		{
			String term = getNextTerm (expr);
			expr = stripNextTerm (expr);
			
			// previous term's result is this term's source
			NodeListImpl source = result;
			result = new NodeListImpl (1);
			
			// absolute terms
			if (term.startsWith ("origin"))
			{
				if (origin != null)
					result.appendNode (origin);
			}
			else if (term.startsWith ("root"))
			{
				Node n = doc.getDocumentElement ();
				
				if (n != null)
					result.appendNode (n);
			}
			else if (term.startsWith ("id"))
			{
				Node n = id (doc.getDocumentElement (), 
					Function.extractParam (term, 1));
				
				if (n != null)
					result.appendNode (n);
			}
			else if (term.startsWith ("html"))
			{
				Node n = name (doc.getDocumentElement (), 
					Function.extractParam (term, 1));
				
				if (n != null)
					result.appendNode (n);
			}
			// relative terms; may have multi source Nodes
			// single Node result
			else if (term.startsWith ("attr"))
			{
				int len = source.getLength ();
				for (int i = 0; i < len; i++)
				{
					Node n = attr (source.item (i), 
						Function.extractParam (term, 1));
					
					if (n != null)
						result.appendNode (n);
				}
			}
			
			// NodeList result
			else if (term.startsWith ("string"))
			{
				int len = source.getLength ();
				for (int i = 0; i < len; i++)
				{
					NodeList partial = string (source.item (i), term);
					int partialLen = partial.getLength ();
					
					for (int j = 0; j < partialLen; j++)
					{
						Node n = partial.item (j);
						if (n != null)
							result.appendNode (n);
					}
				}
			}
			else if (term.startsWith ("child"))
			{
				int len = source.getLength ();
				for (int i = 0; i < len; i++)
				{
					NodeList partial = child (source.item (i), term);
					int partialLen = partial.getLength ();
					
					for (int j = 0; j < partialLen; j++)
					{
						Node n = partial.item (j);
						if (n != null)
							result.appendNode (n);
					}
				}
			}
			else if (term.startsWith ("descendant"))
			{
				int len = source.getLength ();
				for (int i = 0; i < len; i++)
				{
					NodeList partial = descendant (source.item (i), term);
					int partialLen = partial.getLength ();
					
					for (int j = 0; j < partialLen; j++)
					{
						Node n = partial.item (j);
						if (n != null)
							result.appendNode (n);
					}
				}
			}
			else if (term.startsWith ("ancestor"))
			{
				int len = source.getLength ();
				for (int i = 0; i < len; i++)
				{
					NodeList partial = ancestor (source.item (i), term);
					int partialLen = partial.getLength ();
					
					for (int j = 0; j < partialLen; j++)
					{
						Node n = partial.item (j);
						if (n != null)
							result.appendNode (n);
					}
				}
			}
			else if (term.startsWith ("preceding"))
			{
				int len = source.getLength ();
				for (int i = 0; i < len; i++)
				{
					NodeList partial = preceding (source.item (i), term);
					int partialLen = partial.getLength ();
					
					for (int j = 0; j < partialLen; j++)
					{
						Node n = partial.item (j);
						if (n != null)
							result.appendNode (n);
					}
				}
			}
			else if (term.startsWith ("following"))
			{
				int len = source.getLength ();
				for (int i = 0; i < len; i++)
				{
					NodeList partial = following (source.item (i), term);
					int partialLen = partial.getLength ();
					
					for (int j = 0; j < partialLen; j++)
					{
						Node n = partial.item (j);
						if (n != null)
							result.appendNode (n);
					}
				}
			}
			else if (term.startsWith ("psibling"))
			{
				int len = source.getLength ();
				for (int i = 0; i < len; i++)
				{
					NodeList partial = psibling (source.item (i), term);
					int partialLen = partial.getLength ();
					
					for (int j = 0; j < partialLen; j++)
					{
						Node n = partial.item (j);
						if (n != null)
							result.appendNode (n);
					}
				}
			}
			else if (term.startsWith ("fsibling"))
			{
				int len = source.getLength ();
				for (int i = 0; i < len; i++)
				{
					NodeList partial = fsibling (source.item (i), term);
					int partialLen = partial.getLength ();
					
					for (int j = 0; j < partialLen; j++)
					{
						Node n = partial.item (j);
						if (n != null)
							result.appendNode (n);
					}
				}
			}
			else if (term.startsWith ("span"))
			{
				int len = source.getLength ();
				for (int i = 0; i < len; i++)
				{
					NodeList partial = span (source.item (i), term);
					int partialLen = partial.getLength ();
					
					for (int j = 0; j < partialLen; j++)
					{
						Node n = partial.item (j);
						if (n != null)
							result.appendNode (n);
					}
				}
			}
		}
		
		return result;
	}
	
	protected Node id (Element e, String target)
	{
		// done
		if (e.getAttribute ("id").equals (target) ||
			e.getAttribute ("ID").equals (target))
				return e;
		
		NodeList l = e.getChildNodes ();
		int len = l.getLength ();
		
		// try each Element child
		for (int i = 0; i < len; i++)
		{
			Node c = l.item (i);
			
			if (c instanceof Element)
			{
				Node r = id ((Element) c, target);
				
				// if not null, we've found it
				if (r != null)
					return r;
			}
		}
		
		// failed
		return null;
	}
	
	protected Node name (Element e, String target)
	{
		// done
		if (e.getAttribute ("name").equals (target) ||
			e.getAttribute ("NAME").equals (target))
				return e;
		
		NodeList l = e.getChildNodes ();
		int len = l.getLength ();
		
		// try each Element child
		for (int i = 0; i < len; i++)
		{
			Node c = l.item (i);
			
			if (c instanceof Element)
			{
				Node r = id ((Element) c, target);
				
				// if not null, we've found it
				if (r != null)
					return r;
			}
		}
		
		// failed
		return null;
	}
	
	protected Node attr (Node c, String target)
	{
		if (!(c instanceof Element))
			return null;
		
		return ((Element) c).getAttributeNode (target);
	}
	
	protected NodeListImpl string (Node c, String term)
	{
		String temp = Function.extractParam (term, 1);
		int index = 1;
		if (temp.equals ("all"))
			index = Integer.MAX_VALUE;
		else if (!temp.equals (""))
			index = Integer.valueOf (temp).intValue ();
		
		String match = Function.extractParam (term, 2);
		// remove quotation marks
		if (match.length () > 2)
			match = match.substring (1, match.length () - 1);
		else
			match = "";
		
		temp = Function.extractParam (term, 3);
		int offset = 1;
		if (temp.equals ("end"))
			offset = match.length ();
		else if (!temp.equals (""))
			offset = Integer.valueOf (temp).intValue ();
		
		temp = Function.extractParam (term, 4);
		int length = 0;
		if (!temp.equals (""))
			length = Integer.valueOf (temp).intValue ();
		if (length == 0)
			length = 1;
		
		int start = 0;
		String source = c.getNodeValue ();
		NodeListImpl target = new NodeListImpl ();
		
		if (index == Integer.MAX_VALUE)
			// select all matches
			while (true)
			{
				start = source.indexOf (match, start);
				if (start < 0)
					return target;
				
				// make sure the base offset and length aren't too long
				int base = start + offset;
				
				if (base > source.length ())
					return target;
				if (base + length > source.length ())
					length = source.length () - base;
				
				target.appendNode (doc.createTextNode (source.substring 
					(base, base + length)));
				
				start++;
			}
		
		if (match.equals (""))
			start = index;
		
		else for (int i = 0; i < index; i++)
		{
			start = source.indexOf (match, start);
			if (start < 0)
				return target;
		}
		
		// make sure the base offset and length aren't too long
		int base = start + offset;
		
		if (base > source.length ())
			return target;
		if (base + length > source.length ())
			length = source.length () - base;
		
		target.appendNode (doc.createTextNode (source.substring 
			(base, base + length)));
		
		return target;
	}
	
	protected NodeListImpl child (Node c, String term)
	{
		String index = Function.extractParam (term, 1);
		String name = Function.extractParam (term, 2);
		int type = nodeType (name);
		
		// get attribute parameters
		int attrLen = 0;
		String[] attrName = null, attrValue = null;
		
		if (type == Node.ELEMENT_NODE)
		{
			attrLen = nodeAttributePairs (term);
			attrName = new String[attrLen];
			attrValue = new String[attrLen];
			
			for (int i = 0; i < attrLen; i++)
			{
				attrName[i] = Function.extractParam (term, i + 3);
				attrValue[i] = Function.extractParam (term, i + 4);
			}
		}
		
		// pos = target position, count = potential targets reached
		int pos = 0, count = 0;
		if (index.equals (""))
			pos = 1;
		else if (!index.equals ("all"))
			pos = Integer.valueOf (index).intValue ();
		
		NodeList source = c.getChildNodes ();
		int len = source.getLength ();
		NodeListImpl target = new NodeListImpl ();
		
		// select children based on type, element name, and position
		if (pos > -1)
			for (int i = 0; i < len; i++)
			{
				Node n = source.item (i);
				
				if (nodeMatches (n, type, name, attrLen, attrName, attrValue))
				{
					count++;
					
					if (pos == count)
					{
						target.appendNode (n);
						break;
					}
					// ignore position; take all of them
					else if (pos == 0)
						target.appendNode (n);
				}
			}
		else // count backwards
			for (int i = len - 1; i >= 0; i--)
			{
				Node n = source.item (i);
				
				if (nodeMatches (n, type, name, attrLen, attrName, attrValue))
				{
					count--;
					
					if (pos == count)
					{
						target.appendNode (n);
						break;
					}
				}
			}
		
		return target;
	}
	
	protected NodeListImpl descendant (Node c, String term)
	{
		String index = Function.extractParam (term, 1);
		String name = Function.extractParam (term, 2);
		int type = nodeType (name);
		
		// get attribute parameters
		int attrLen = 0;
		String[] attrName = null, attrValue = null;
		
		if (type == Node.ELEMENT_NODE)
		{
			attrLen = nodeAttributePairs (term);
			attrName = new String[attrLen];
			attrValue = new String[attrLen];
			
			for (int i = 0; i < attrLen; i++)
			{
				attrName[i] = Function.extractParam (term, i + 3);
				attrValue[i] = Function.extractParam (term, i + 4);
			}
		}
		
		// pos = target position, count = potential targets reached
		VolatileInteger count = new VolatileInteger (0);
		int pos = 0;
		if (index.equals (""))
			pos = 1;
		else if (!index.equals ("all"))
			pos = Integer.valueOf (index).intValue ();
		
		return descendant (c, type, name, attrLen, attrName, attrValue, 
			pos, count);
	}
	
	protected NodeListImpl descendant (Node c, int type, String name, 
		int attrLen, String[] attrName, String[] attrValue, 
		int pos, VolatileInteger count)
	{
		NodeList source = c.getChildNodes ();
		int len = source.getLength ();
		NodeListImpl target = new NodeListImpl ();
		
		// select children based on type, element name, and position
		if (pos > -1)
			for (int i = 0; i < len; i++)
			{
				Node n = source.item (i);
				
				if (nodeMatches (n, type, name, attrLen, attrName, attrValue))
				{
					count.increment ();
					
					if (count.equals (pos))
					{
						target.appendNode (n);
						break;
					}
					// ignore position; take all of them
					else if (pos == 0)
						target.appendNode (n);
				}
				
				// copy results to target NodeList
				NodeList desTarget = descendant (n, type, name, attrLen, 
					attrName, attrValue, pos, count);
				int desLen = desTarget.getLength ();
				
				for (int j = 0; j < desLen; j++)
					target.appendNode (desTarget.item (j));
			}
		else // count backwards
			for (int i = len - 1; i >= 0; i--)
			{
				Node n = source.item (i);
				
				// if we've found the right node already, break out and return
				target = descendant (n, type, name, attrLen, attrName, attrValue,
					pos, count);
				if (target.getLength () > 0) break;
				
				if (nodeMatches (n, type, name, attrLen, attrName, attrValue))
				{
					count.decrement ();
					
					if (count.equals (pos))
					{
						target.appendNode (n);
						break;
					}
				}
			}
		
		return target;
	}
	
	protected NodeListImpl ancestor (Node c, String term)
	{
		String index = Function.extractParam (term, 1);
		String name = Function.extractParam (term, 2);
		int type = nodeType (name);
		
		// get attribute parameters
		int attrLen = 0;
		String[] attrName = null, attrValue = null;
		
		if (type == Node.ELEMENT_NODE)
		{
			attrLen = nodeAttributePairs (term);
			attrName = new String[attrLen];
			attrValue = new String[attrLen];
			
			for (int i = 0; i < attrLen; i++)
			{
				attrName[i] = Function.extractParam (term, i + 3);
				attrValue[i] = Function.extractParam (term, i + 4);
			}
		}
		
		// pos = target position, count = potential targets reached
		int pos = 0, count = 0;
		if (index.equals (""))
			pos = 1;
		else if (!index.equals ("all"))
			pos = Integer.valueOf (index).intValue ();
		
		Element root = doc.getDocumentElement ();
		Node n = c.getParentNode ();
		NodeListImpl target = new NodeListImpl ();
		
		// select ancestors based on type, element name, and position
		if (pos > -1)
			while (n != root)
			{
				if (nodeMatches (n, type, name, attrLen, attrName, attrValue))
				{
					count++;
					
					if (pos == count)
					{
						target.appendNode (n);
						break;
					}
					// ignore position; take all of them
					else if (pos == 0)
						target.appendNode (n);
				}
				
				n = n.getParentNode ();
			}
		else // gather ancestors, then step through them backwards (top down)
		{
			NodeListImpl source = new NodeListImpl ();
			while (n != root)
			{
				source.appendNode (n);
				n = n.getParentNode ();
			}
			
			int len = source.getLength ();
			for (int i = len - 1; i >= 0; i--)
			{
				n = source.item (i);
				
				if (nodeMatches (n, type, name, attrLen, attrName, attrValue))
				{
					count--;
					
					if (pos == count)
					{
						target.appendNode (n);
						break;
					}
				}
			}
		}
		
		return target;
	}
	
	protected NodeListImpl preceding (Node c, String term)
	{
		String index = Function.extractParam (term, 1);
		String name = Function.extractParam (term, 2);
		int type = nodeType (name);
		
		// get attribute parameters
		int attrLen = 0;
		String[] attrName = null, attrValue = null;
		
		if (type == Node.ELEMENT_NODE)
		{
			attrLen = nodeAttributePairs (term);
			attrName = new String[attrLen];
			attrValue = new String[attrLen];
			
			for (int i = 0; i < attrLen; i++)
			{
				attrName[i] = Function.extractParam (term, i + 3);
				attrValue[i] = Function.extractParam (term, i + 4);
			}
		}
		
		// pos = target position, count = potential targets reached
		VolatileInteger count = new VolatileInteger (0);
		int pos = 0;
		if (index.equals (""))
			pos = 1;
		else if (!index.equals ("all"))
			pos = Integer.valueOf (index).intValue ();
		
		return preceding ((pos > -1 ? c.getParentNode () : doc.getDocumentElement ()), 
			type, name, attrLen, attrName, attrValue, pos, count, c);
	}
	
	protected NodeListImpl preceding (Node c, int type, String name, 
		int attrLen, String[] attrName, String[] attrValue, 
		int pos, VolatileInteger count, Node end)
	{
		NodeListImpl target = new NodeListImpl ();
		NodeList source = c.getChildNodes ();
		int len = source.getLength ();
		
		Node n = null;
		int inFront = 0;
		
		// select children in front of end, then the parent node
		if (pos > -1)
		{
			for (inFront = 0; inFront < len; inFront++)
				if (source.item (inFront) == end)
					break;
			
			for (int i = inFront - 1; i >= 0; i--)
			{
				n = source.item (i);
				
				// try children first
				NodeList desTarg = descendant (n, type, name, attrLen, 
					attrName, attrValue, pos, count);
				int desLen = desTarg.getLength ();
				
				for (int j = 0; j < desLen; j++)
					target.appendNode (desTarg.item (j));
				
				// quit if we've found the specified position
				if (pos > 0 && desLen > 0)
					break;
				
				if (nodeMatches (n, type, name, attrLen, attrName, attrValue))
				{
					count.increment ();
					
					if (count.equals (pos))
					{
						target.appendNode (n);
						break;
					}
					// ignore position; take all of them
					else if (pos == 0)
						target.appendNode (n);
				}
			}
			
			// can't look up if we've reached the root
			if (c != doc.getDocumentElement ())
			{
				// this element now marks the end of the children of its parent
				NodeList preTarg = preceding (c.getParentNode (), type, name, 
					attrLen, attrName, attrValue, pos, count, c);
				int preLen = preTarg.getLength ();
				
				for (int j = 0; j < preLen; j++)
					target.appendNode (preTarg.item (j));
			}
		}
		else // count backwards (top down)
			for (int i = 0; i < len; i++)
			{
				n = source.item (i);
				if (n == end)
					break;
				
				if (nodeMatches (n, type, name, attrLen, attrName, attrValue))
				{
					count.decrement ();
					
					if (count.equals (pos))
					{
						target.appendNode (n);
						break;
					}
				}
				
				target = preceding (n, type, name, attrLen, attrName, attrValue, 
					pos, count, end);
			}
		
		return target;
	}
	
	protected NodeListImpl following (Node c, String term)
	{
		String index = Function.extractParam (term, 1);
		String name = Function.extractParam (term, 2);
		int type = nodeType (name);
		
		// get attribute parameters
		int attrLen = 0;
		String[] attrName = null, attrValue = null;
		
		if (type == Node.ELEMENT_NODE)
		{
			attrLen = nodeAttributePairs (term);
			attrName = new String[attrLen];
			attrValue = new String[attrLen];
			
			for (int i = 0; i < attrLen; i++)
			{
				attrName[i] = Function.extractParam (term, i + 3);
				attrValue[i] = Function.extractParam (term, i + 4);
			}
		}
		
		// pos = target position, count = potential targets reached
		VolatileInteger count = new VolatileInteger (0);
		int pos = 0;
		if (index.equals (""))
			pos = 1;
		else if (!index.equals ("all"))
			pos = Integer.valueOf (index).intValue ();
		
		return following ((pos > -1 ? c.getParentNode () : doc.getDocumentElement ()), 
			type, name, attrLen, attrName, attrValue, pos, count, c);
	}
	
	protected NodeListImpl following (Node c, int type, String name, 
		int attrLen, String[] attrName, String[] attrValue, 
		int pos, VolatileInteger count, Node end)
	{
		NodeListImpl target = new NodeListImpl ();
		NodeList source = c.getChildNodes ();
		int len = source.getLength ();
		
		Node n = null;
		int inFront = 0;
		
		// select children behind the end
		if (pos > -1)
		{
			for (inFront = 0; inFront < len; inFront++)
				if (source.item (inFront) == end)
					break;
			
			for (int i = inFront + 1; i < len; i++)
			{
				n = source.item (i);
				
				// try children (but *not* this Node c)
				NodeList desTarg = descendant (n, type, name, attrLen, attrName, 
					attrValue, pos, count);
				int desLen = desTarg.getLength ();
				
				for (int j = 0; j < desLen; j++)
					target.appendNode (desTarg.item (j));
				
				// quit if we've found the specified position
				if (pos > 0 && desLen > 0)
					break;
			}
			
			// can't look up if we've reached the root
			if (c == doc.getDocumentElement ())
			{
				// this element now marks the end of the children of its parent
				NodeList folTarg = following (c.getParentNode (), type, name, 
					attrLen, attrName, attrValue, pos, count, c);
				int folLen = folTarg.getLength ();
				
				for (int j = 0; j < folLen; j++)
					target.appendNode (folTarg.item (j));
			}
		}
		else // count backwards (bottom up)
			for (int i = len - 1; i >= 0; i--)
			{
				n = source.item (i);
				
				// end is flag; indicates no matches were found
				// will cause ancestors of end to be skipped
				if (n == end)
				{
					target.appendNode (end);
					break;
				}
				
				target = following (n, type, name, attrLen, attrName, attrValue, 
					pos, count, end);
				if (target.getLength () > 0)
					break;			
				
				if (nodeMatches (n, type, name, attrLen, attrName, attrValue))
				{
					count.decrement ();
					
					if (count.equals (pos))
					{
						target.appendNode (n);
						break;
					}
				}
				
			}
		
		// if true, no targets were actually found
		if (target.getLength () == 1 && target.item (0) == end)
			target.clearNodes ();
		
		return target;
	}
	
	protected NodeListImpl psibling (Node c, String term)
	{
		String index = Function.extractParam (term, 1);
		String name = Function.extractParam (term, 2);
		int type = nodeType (name);
		
		// get attribute parameters
		int attrLen = 0;
		String[] attrName = null, attrValue = null;
		
		if (type == Node.ELEMENT_NODE)
		{
			attrLen = nodeAttributePairs (term);
			attrName = new String[attrLen];
			attrValue = new String[attrLen];
			
			for (int i = 0; i < attrLen; i++)
			{
				attrName[i] = Function.extractParam (term, i + 3);
				attrValue[i] = Function.extractParam (term, i + 4);
			}
		}
		
		// pos = target position, count = potential targets reached
		int pos = 0, count = 0;
		if (index.equals (""))
			pos = 1;
		else if (!index.equals ("all"))
			pos = Integer.valueOf (index).intValue ();
		
		NodeList source = c.getParentNode ().getChildNodes ();
		int len = source.getLength ();
		NodeListImpl target = new NodeListImpl ();
		
		Node n = null;
		int inFront = 0;
		
		// select children based on type, element name, and position
		if (pos > -1)
		{
			for (inFront = 0; inFront < len; inFront++)
				if (source.item (inFront) == c)
					break;
			
			for (int i = inFront - 1; i >= 0; i--)
			{
				n = source.item (i);
				
				if (nodeMatches (n, type, name, attrLen, attrName, attrValue))
				{
					count++;
					
					if (pos == count)
					{
						target.appendNode (n);
						break;
					}
					// ignore position; take all of them
					else if (pos == 0)
						target.appendNode (n);
				}
			}
		}
		else // count backwards
			for (int i = 0; i < len; i++)
			{
				n = source.item (i);
				if (n == c)
					break;
				
				if (nodeMatches (n, type, name, attrLen, attrName, attrValue))
				{
					count--;
					
					if (pos == count)
					{
						target.appendNode (n);
						break;
					}
				}
			}
		
		return target;
	}
	
	protected NodeListImpl fsibling (Node c, String term)
	{
		String index = Function.extractParam (term, 1);
		String name = Function.extractParam (term, 2);
		int type = nodeType (name);
		
		// get attribute parameters
		int attrLen = 0;
		String[] attrName = null, attrValue = null;
		
		if (type == Node.ELEMENT_NODE)
		{
			attrLen = nodeAttributePairs (term);
			attrName = new String[attrLen];
			attrValue = new String[attrLen];
			
			for (int i = 0; i < attrLen; i++)
			{
				attrName[i] = Function.extractParam (term, i + 3);
				attrValue[i] = Function.extractParam (term, i + 4);
			}
		}
		
		// pos = target position, count = potential targets reached
		int pos = 0, count = 0;
		if (index.equals (""))
			pos = 1;
		else if (!index.equals ("all"))
			pos = Integer.valueOf (index).intValue ();
		
		NodeList source = c.getParentNode ().getChildNodes ();
		int len = source.getLength ();
		NodeListImpl target = new NodeListImpl ();
		
		Node n = null;
		int inFront = 0;
		
		// select children based on type, element name, and position
		if (pos > -1)
		{
			for (inFront = 0; inFront < len && n != c; inFront++)
				n = source.item (inFront);
			
			for (int i = inFront + 1; i < len; i++)
			{
				n = source.item (i);
				
				if (nodeMatches (n, type, name, attrLen, attrName, attrValue))
				{
					count++;
					
					if (pos == count)
					{
						target.appendNode (n);
						break;
					}
					// ignore position; take all of them
					else if (pos == 0)
						target.appendNode (n);
				}
			}
		}
		else // count backwards
			for (int i = len - 1; i >= 0; i--)
			{
				n = source.item (i);
				if (n == c)
					break;
				
				if (nodeMatches (n, type, name, attrLen, attrName, attrValue))
				{
					count--;
					
					if (pos == count)
					{
						target.appendNode (n);
						break;
					}
				}
			}
		
		return target;
	}
	
	protected NodeListImpl span (Node c, String term)
	{
		int start = 0, middle = 0, end = 0, 
			parens = 0, index = 0, len = term.length ();
		
		start = term.indexOf ('(') + 1;
		end = term.lastIndexOf (')');
		
		for (index = 0; index < len; index++)
		{
			int ch = term.indexOf (index);
			
			if (ch == ',' && parens < 1)
			{
				middle = index;
				break;
			}
			else if (ch == '(')
				parens++;
			else if (ch == ')')
				parens--;
		}
		
		NodeListImpl one = resolve (c, term.substring (start, middle).trim ());
		NodeListImpl two = resolve (c, term.substring (middle + 1, end).trim ());
		
		if (one.getLength () < 1)
			return two;
		if (two.getLength () < 1)
			return one;
		
		Node first = one.item (0);
		Node last = two.item (0);
		NodeListImpl target = new NodeListImpl (10);
		
		// look for last element as child of first
		if (!span (first, last, target))
			
			// then look for it as one of first's fsibling,
			// and then as an fsibling of first's ancestors
			while (first != doc.getDocumentElement ())
			{
				Node next = first.getNextSibling ();
				
				while (next != null)
				{
					if (span (next, last, target))
						return target;
					
					next = next.getNextSibling ();
				}
				
				first = first.getParentNode ();
			}
		
		return target;
	}
	
	protected boolean span (Node first, Node last, NodeListImpl target)
	{
		// add current node
		target.appendNode (first);
		
		if (first == last)
			return true;
		
		// add child nodes until end is found
		NodeList children = first.getChildNodes ();
		int len = children.getLength ();
		
		for (int i = 0; i < len; i++)
			if (span (children.item (i), last, target))
				return true;
		
		return false;
	}
	
	// type = -1 if name == "#all"
	protected int nodeType (String name)
	{
		int type = -1;
		
		if (name.equals (""))
			;
		else if (!name.startsWith ("#") || name.equals ("#element"))
			type = Node.ELEMENT_NODE;
		else if (name.equals ("#pi"))
			type = Node.PROCESSING_INSTRUCTION_NODE;
		else if (name.equals ("#text"))
			type = Node.TEXT_NODE;
		else if (name.equals ("#cdata"))
			type = Node.CDATA_SECTION_NODE;
		else if (name.equals ("#comment"))
			type = Node.COMMENT_NODE;
		
		return type;
	}
	
	// returns the number of attribute pairs in the term parameter list
	protected int nodeAttributePairs (String term)
	{
		int comma = 0;
		int count = 0;
		
		while (comma > -1)
		{
			comma = term.indexOf (',', comma + 1);
			count++;
		}
		
		return (count - 1) / 2;
	}
	
	// type == -1 if type == #all
	protected boolean nodeMatches (Node n, int type, String name, 
				int attrLen, String[] attrName, String[] attrValue)
	{
		boolean matches = false;
		
		// if name is a specific element name, make sure it matches
		if ((type == -1 || n.getNodeType () == type) &&
			(name.startsWith ("#") || n.getNodeName ().equals (name) 
			|| name.equals("")))
		{
			matches = true;
			
			if (n instanceof Element)
			{
				Element e = (Element) n;
				
				// check each attribute parameter name & value pair
				for (int j = 0; j < attrLen; j++)
				{
					if (attrName[j].equals ("*"))
					{
						NamedNodeMap map = e.getAttributes ();
						int k, mapLen = map.getLength ();
						
						// check every attribute
						// for a value matching the value param
						for (k = 0; k < mapLen; k++)
							if (((Attr) map.item (k)).getValue ().equals 
								(attrValue[j]))
								break;
						
						if (k == mapLen)
						{
							matches = false;
							break;
						}
							
					}
					else if (attrValue[j].equals ("*") && 
							e.getAttributeNode (attrName[j]) == null)
					{
						matches = false;
						break;
					}
					else if (!e.getAttribute (attrName[j]).equals (attrValue[j]))
					{
						matches = false;
						break;
					}
				}
			}
		}
		
		return matches;
	}
}